#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file

Licensed to the Apache Software Foundation (ASF) under one or more
contributor license agreements.  See the NOTICE file distributed with
this work for additional information regarding copyright ownership.
The ASF licenses this file to You under the Apache License, Version 2.0
(the "License"); you may not use this file except in compliance with
the License.  You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import builtins
import collections.abc
import google.protobuf.any_pb2
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import pyspark.sql.connect.proto.common_pb2
import pyspark.sql.connect.proto.expressions_pb2
import pyspark.sql.connect.proto.ml_pb2
import pyspark.sql.connect.proto.pipelines_pb2
import pyspark.sql.connect.proto.relations_pb2
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _StreamingQueryEventType:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _StreamingQueryEventTypeEnumTypeWrapper(
    google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_StreamingQueryEventType.ValueType],
    builtins.type,
):  # noqa: F821
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    QUERY_PROGRESS_UNSPECIFIED: _StreamingQueryEventType.ValueType  # 0
    QUERY_PROGRESS_EVENT: _StreamingQueryEventType.ValueType  # 1
    QUERY_TERMINATED_EVENT: _StreamingQueryEventType.ValueType  # 2
    QUERY_IDLE_EVENT: _StreamingQueryEventType.ValueType  # 3

class StreamingQueryEventType(
    _StreamingQueryEventType, metaclass=_StreamingQueryEventTypeEnumTypeWrapper
):
    """The enum used for client side streaming query listener event
    There is no QueryStartedEvent defined here,
    it is added as a field in WriteStreamOperationStartResult
    """

QUERY_PROGRESS_UNSPECIFIED: StreamingQueryEventType.ValueType  # 0
QUERY_PROGRESS_EVENT: StreamingQueryEventType.ValueType  # 1
QUERY_TERMINATED_EVENT: StreamingQueryEventType.ValueType  # 2
QUERY_IDLE_EVENT: StreamingQueryEventType.ValueType  # 3
global___StreamingQueryEventType = StreamingQueryEventType

class Command(google.protobuf.message.Message):
    """A [[Command]] is an operation that is executed by the server that does not directly consume or
    produce a relational result.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    REGISTER_FUNCTION_FIELD_NUMBER: builtins.int
    WRITE_OPERATION_FIELD_NUMBER: builtins.int
    CREATE_DATAFRAME_VIEW_FIELD_NUMBER: builtins.int
    WRITE_OPERATION_V2_FIELD_NUMBER: builtins.int
    SQL_COMMAND_FIELD_NUMBER: builtins.int
    WRITE_STREAM_OPERATION_START_FIELD_NUMBER: builtins.int
    STREAMING_QUERY_COMMAND_FIELD_NUMBER: builtins.int
    GET_RESOURCES_COMMAND_FIELD_NUMBER: builtins.int
    STREAMING_QUERY_MANAGER_COMMAND_FIELD_NUMBER: builtins.int
    REGISTER_TABLE_FUNCTION_FIELD_NUMBER: builtins.int
    STREAMING_QUERY_LISTENER_BUS_COMMAND_FIELD_NUMBER: builtins.int
    REGISTER_DATA_SOURCE_FIELD_NUMBER: builtins.int
    CREATE_RESOURCE_PROFILE_COMMAND_FIELD_NUMBER: builtins.int
    CHECKPOINT_COMMAND_FIELD_NUMBER: builtins.int
    REMOVE_CACHED_REMOTE_RELATION_COMMAND_FIELD_NUMBER: builtins.int
    MERGE_INTO_TABLE_COMMAND_FIELD_NUMBER: builtins.int
    ML_COMMAND_FIELD_NUMBER: builtins.int
    EXECUTE_EXTERNAL_COMMAND_FIELD_NUMBER: builtins.int
    PIPELINE_COMMAND_FIELD_NUMBER: builtins.int
    EXTENSION_FIELD_NUMBER: builtins.int
    @property
    def register_function(
        self,
    ) -> pyspark.sql.connect.proto.expressions_pb2.CommonInlineUserDefinedFunction: ...
    @property
    def write_operation(self) -> global___WriteOperation: ...
    @property
    def create_dataframe_view(self) -> global___CreateDataFrameViewCommand: ...
    @property
    def write_operation_v2(self) -> global___WriteOperationV2: ...
    @property
    def sql_command(self) -> global___SqlCommand: ...
    @property
    def write_stream_operation_start(self) -> global___WriteStreamOperationStart: ...
    @property
    def streaming_query_command(self) -> global___StreamingQueryCommand: ...
    @property
    def get_resources_command(self) -> global___GetResourcesCommand: ...
    @property
    def streaming_query_manager_command(self) -> global___StreamingQueryManagerCommand: ...
    @property
    def register_table_function(
        self,
    ) -> pyspark.sql.connect.proto.relations_pb2.CommonInlineUserDefinedTableFunction: ...
    @property
    def streaming_query_listener_bus_command(self) -> global___StreamingQueryListenerBusCommand: ...
    @property
    def register_data_source(
        self,
    ) -> pyspark.sql.connect.proto.relations_pb2.CommonInlineUserDefinedDataSource: ...
    @property
    def create_resource_profile_command(self) -> global___CreateResourceProfileCommand: ...
    @property
    def checkpoint_command(self) -> global___CheckpointCommand: ...
    @property
    def remove_cached_remote_relation_command(
        self,
    ) -> global___RemoveCachedRemoteRelationCommand: ...
    @property
    def merge_into_table_command(self) -> global___MergeIntoTableCommand: ...
    @property
    def ml_command(self) -> pyspark.sql.connect.proto.ml_pb2.MlCommand: ...
    @property
    def execute_external_command(self) -> global___ExecuteExternalCommand: ...
    @property
    def pipeline_command(self) -> pyspark.sql.connect.proto.pipelines_pb2.PipelineCommand: ...
    @property
    def extension(self) -> google.protobuf.any_pb2.Any:
        """This field is used to mark extensions to the protocol. When plugins generate arbitrary
        Commands they can add them here. During the planning the correct resolution is done.
        """
    def __init__(
        self,
        *,
        register_function: pyspark.sql.connect.proto.expressions_pb2.CommonInlineUserDefinedFunction
        | None = ...,
        write_operation: global___WriteOperation | None = ...,
        create_dataframe_view: global___CreateDataFrameViewCommand | None = ...,
        write_operation_v2: global___WriteOperationV2 | None = ...,
        sql_command: global___SqlCommand | None = ...,
        write_stream_operation_start: global___WriteStreamOperationStart | None = ...,
        streaming_query_command: global___StreamingQueryCommand | None = ...,
        get_resources_command: global___GetResourcesCommand | None = ...,
        streaming_query_manager_command: global___StreamingQueryManagerCommand | None = ...,
        register_table_function: pyspark.sql.connect.proto.relations_pb2.CommonInlineUserDefinedTableFunction
        | None = ...,
        streaming_query_listener_bus_command: global___StreamingQueryListenerBusCommand
        | None = ...,
        register_data_source: pyspark.sql.connect.proto.relations_pb2.CommonInlineUserDefinedDataSource
        | None = ...,
        create_resource_profile_command: global___CreateResourceProfileCommand | None = ...,
        checkpoint_command: global___CheckpointCommand | None = ...,
        remove_cached_remote_relation_command: global___RemoveCachedRemoteRelationCommand
        | None = ...,
        merge_into_table_command: global___MergeIntoTableCommand | None = ...,
        ml_command: pyspark.sql.connect.proto.ml_pb2.MlCommand | None = ...,
        execute_external_command: global___ExecuteExternalCommand | None = ...,
        pipeline_command: pyspark.sql.connect.proto.pipelines_pb2.PipelineCommand | None = ...,
        extension: google.protobuf.any_pb2.Any | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "checkpoint_command",
            b"checkpoint_command",
            "command_type",
            b"command_type",
            "create_dataframe_view",
            b"create_dataframe_view",
            "create_resource_profile_command",
            b"create_resource_profile_command",
            "execute_external_command",
            b"execute_external_command",
            "extension",
            b"extension",
            "get_resources_command",
            b"get_resources_command",
            "merge_into_table_command",
            b"merge_into_table_command",
            "ml_command",
            b"ml_command",
            "pipeline_command",
            b"pipeline_command",
            "register_data_source",
            b"register_data_source",
            "register_function",
            b"register_function",
            "register_table_function",
            b"register_table_function",
            "remove_cached_remote_relation_command",
            b"remove_cached_remote_relation_command",
            "sql_command",
            b"sql_command",
            "streaming_query_command",
            b"streaming_query_command",
            "streaming_query_listener_bus_command",
            b"streaming_query_listener_bus_command",
            "streaming_query_manager_command",
            b"streaming_query_manager_command",
            "write_operation",
            b"write_operation",
            "write_operation_v2",
            b"write_operation_v2",
            "write_stream_operation_start",
            b"write_stream_operation_start",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "checkpoint_command",
            b"checkpoint_command",
            "command_type",
            b"command_type",
            "create_dataframe_view",
            b"create_dataframe_view",
            "create_resource_profile_command",
            b"create_resource_profile_command",
            "execute_external_command",
            b"execute_external_command",
            "extension",
            b"extension",
            "get_resources_command",
            b"get_resources_command",
            "merge_into_table_command",
            b"merge_into_table_command",
            "ml_command",
            b"ml_command",
            "pipeline_command",
            b"pipeline_command",
            "register_data_source",
            b"register_data_source",
            "register_function",
            b"register_function",
            "register_table_function",
            b"register_table_function",
            "remove_cached_remote_relation_command",
            b"remove_cached_remote_relation_command",
            "sql_command",
            b"sql_command",
            "streaming_query_command",
            b"streaming_query_command",
            "streaming_query_listener_bus_command",
            b"streaming_query_listener_bus_command",
            "streaming_query_manager_command",
            b"streaming_query_manager_command",
            "write_operation",
            b"write_operation",
            "write_operation_v2",
            b"write_operation_v2",
            "write_stream_operation_start",
            b"write_stream_operation_start",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["command_type", b"command_type"]
    ) -> (
        typing_extensions.Literal[
            "register_function",
            "write_operation",
            "create_dataframe_view",
            "write_operation_v2",
            "sql_command",
            "write_stream_operation_start",
            "streaming_query_command",
            "get_resources_command",
            "streaming_query_manager_command",
            "register_table_function",
            "streaming_query_listener_bus_command",
            "register_data_source",
            "create_resource_profile_command",
            "checkpoint_command",
            "remove_cached_remote_relation_command",
            "merge_into_table_command",
            "ml_command",
            "execute_external_command",
            "pipeline_command",
            "extension",
        ]
        | None
    ): ...

global___Command = Command

class SqlCommand(google.protobuf.message.Message):
    """A SQL Command is used to trigger the eager evaluation of SQL commands in Spark.

    When the SQL provide as part of the message is a command it will be immediately evaluated
    and the result will be collected and returned as part of a LocalRelation. If the result is
    not a command, the operation will simply return a SQL Relation. This allows the client to be
    almost oblivious to the server-side behavior.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class ArgsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> pyspark.sql.connect.proto.expressions_pb2.Expression.Literal: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: pyspark.sql.connect.proto.expressions_pb2.Expression.Literal | None = ...,
        ) -> None: ...
        def HasField(
            self, field_name: typing_extensions.Literal["value", b"value"]
        ) -> builtins.bool: ...
        def ClearField(
            self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]
        ) -> None: ...

    class NamedArgumentsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> pyspark.sql.connect.proto.expressions_pb2.Expression: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: pyspark.sql.connect.proto.expressions_pb2.Expression | None = ...,
        ) -> None: ...
        def HasField(
            self, field_name: typing_extensions.Literal["value", b"value"]
        ) -> builtins.bool: ...
        def ClearField(
            self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]
        ) -> None: ...

    SQL_FIELD_NUMBER: builtins.int
    ARGS_FIELD_NUMBER: builtins.int
    POS_ARGS_FIELD_NUMBER: builtins.int
    NAMED_ARGUMENTS_FIELD_NUMBER: builtins.int
    POS_ARGUMENTS_FIELD_NUMBER: builtins.int
    INPUT_FIELD_NUMBER: builtins.int
    sql: builtins.str
    """(Required) SQL Query."""
    @property
    def args(
        self,
    ) -> google.protobuf.internal.containers.MessageMap[
        builtins.str, pyspark.sql.connect.proto.expressions_pb2.Expression.Literal
    ]:
        """(Optional) A map of parameter names to literal expressions."""
    @property
    def pos_args(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
        pyspark.sql.connect.proto.expressions_pb2.Expression.Literal
    ]:
        """(Optional) A sequence of literal expressions for positional parameters in the SQL query text."""
    @property
    def named_arguments(
        self,
    ) -> google.protobuf.internal.containers.MessageMap[
        builtins.str, pyspark.sql.connect.proto.expressions_pb2.Expression
    ]:
        """(Optional) A map of parameter names to expressions.
        It cannot coexist with `pos_arguments`.
        """
    @property
    def pos_arguments(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
        pyspark.sql.connect.proto.expressions_pb2.Expression
    ]:
        """(Optional) A sequence of expressions for positional parameters in the SQL query text.
        It cannot coexist with `named_arguments`.
        """
    @property
    def input(self) -> pyspark.sql.connect.proto.relations_pb2.Relation:
        """(Optional) The relation that this SQL command will be built on."""
    def __init__(
        self,
        *,
        sql: builtins.str = ...,
        args: collections.abc.Mapping[
            builtins.str, pyspark.sql.connect.proto.expressions_pb2.Expression.Literal
        ]
        | None = ...,
        pos_args: collections.abc.Iterable[
            pyspark.sql.connect.proto.expressions_pb2.Expression.Literal
        ]
        | None = ...,
        named_arguments: collections.abc.Mapping[
            builtins.str, pyspark.sql.connect.proto.expressions_pb2.Expression
        ]
        | None = ...,
        pos_arguments: collections.abc.Iterable[
            pyspark.sql.connect.proto.expressions_pb2.Expression
        ]
        | None = ...,
        input: pyspark.sql.connect.proto.relations_pb2.Relation | None = ...,
    ) -> None: ...
    def HasField(
        self, field_name: typing_extensions.Literal["input", b"input"]
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "args",
            b"args",
            "input",
            b"input",
            "named_arguments",
            b"named_arguments",
            "pos_args",
            b"pos_args",
            "pos_arguments",
            b"pos_arguments",
            "sql",
            b"sql",
        ],
    ) -> None: ...

global___SqlCommand = SqlCommand

class CreateDataFrameViewCommand(google.protobuf.message.Message):
    """A command that can create DataFrame global temp view or local temp view."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    INPUT_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    IS_GLOBAL_FIELD_NUMBER: builtins.int
    REPLACE_FIELD_NUMBER: builtins.int
    @property
    def input(self) -> pyspark.sql.connect.proto.relations_pb2.Relation:
        """(Required) The relation that this view will be built on."""
    name: builtins.str
    """(Required) View name."""
    is_global: builtins.bool
    """(Required) Whether this is global temp view or local temp view."""
    replace: builtins.bool
    """(Required)

    If true, and if the view already exists, updates it; if false, and if the view
    already exists, throws exception.
    """
    def __init__(
        self,
        *,
        input: pyspark.sql.connect.proto.relations_pb2.Relation | None = ...,
        name: builtins.str = ...,
        is_global: builtins.bool = ...,
        replace: builtins.bool = ...,
    ) -> None: ...
    def HasField(
        self, field_name: typing_extensions.Literal["input", b"input"]
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "input", b"input", "is_global", b"is_global", "name", b"name", "replace", b"replace"
        ],
    ) -> None: ...

global___CreateDataFrameViewCommand = CreateDataFrameViewCommand

class WriteOperation(google.protobuf.message.Message):
    """As writes are not directly handled during analysis and planning, they are modeled as commands."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _SaveMode:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _SaveModeEnumTypeWrapper(
        google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[
            WriteOperation._SaveMode.ValueType
        ],
        builtins.type,
    ):  # noqa: F821
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        SAVE_MODE_UNSPECIFIED: WriteOperation._SaveMode.ValueType  # 0
        SAVE_MODE_APPEND: WriteOperation._SaveMode.ValueType  # 1
        SAVE_MODE_OVERWRITE: WriteOperation._SaveMode.ValueType  # 2
        SAVE_MODE_ERROR_IF_EXISTS: WriteOperation._SaveMode.ValueType  # 3
        SAVE_MODE_IGNORE: WriteOperation._SaveMode.ValueType  # 4

    class SaveMode(_SaveMode, metaclass=_SaveModeEnumTypeWrapper): ...
    SAVE_MODE_UNSPECIFIED: WriteOperation.SaveMode.ValueType  # 0
    SAVE_MODE_APPEND: WriteOperation.SaveMode.ValueType  # 1
    SAVE_MODE_OVERWRITE: WriteOperation.SaveMode.ValueType  # 2
    SAVE_MODE_ERROR_IF_EXISTS: WriteOperation.SaveMode.ValueType  # 3
    SAVE_MODE_IGNORE: WriteOperation.SaveMode.ValueType  # 4

    class OptionsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(
            self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]
        ) -> None: ...

    class SaveTable(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        class _TableSaveMethod:
            ValueType = typing.NewType("ValueType", builtins.int)
            V: typing_extensions.TypeAlias = ValueType

        class _TableSaveMethodEnumTypeWrapper(
            google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[
                WriteOperation.SaveTable._TableSaveMethod.ValueType
            ],
            builtins.type,
        ):  # noqa: F821
            DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
            TABLE_SAVE_METHOD_UNSPECIFIED: WriteOperation.SaveTable._TableSaveMethod.ValueType  # 0
            TABLE_SAVE_METHOD_SAVE_AS_TABLE: WriteOperation.SaveTable._TableSaveMethod.ValueType  # 1
            TABLE_SAVE_METHOD_INSERT_INTO: WriteOperation.SaveTable._TableSaveMethod.ValueType  # 2

        class TableSaveMethod(_TableSaveMethod, metaclass=_TableSaveMethodEnumTypeWrapper): ...
        TABLE_SAVE_METHOD_UNSPECIFIED: WriteOperation.SaveTable.TableSaveMethod.ValueType  # 0
        TABLE_SAVE_METHOD_SAVE_AS_TABLE: WriteOperation.SaveTable.TableSaveMethod.ValueType  # 1
        TABLE_SAVE_METHOD_INSERT_INTO: WriteOperation.SaveTable.TableSaveMethod.ValueType  # 2

        TABLE_NAME_FIELD_NUMBER: builtins.int
        SAVE_METHOD_FIELD_NUMBER: builtins.int
        table_name: builtins.str
        """(Required) The table name."""
        save_method: global___WriteOperation.SaveTable.TableSaveMethod.ValueType
        """(Required) The method to be called to write to the table."""
        def __init__(
            self,
            *,
            table_name: builtins.str = ...,
            save_method: global___WriteOperation.SaveTable.TableSaveMethod.ValueType = ...,
        ) -> None: ...
        def ClearField(
            self,
            field_name: typing_extensions.Literal[
                "save_method", b"save_method", "table_name", b"table_name"
            ],
        ) -> None: ...

    class BucketBy(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        BUCKET_COLUMN_NAMES_FIELD_NUMBER: builtins.int
        NUM_BUCKETS_FIELD_NUMBER: builtins.int
        @property
        def bucket_column_names(
            self,
        ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
        num_buckets: builtins.int
        def __init__(
            self,
            *,
            bucket_column_names: collections.abc.Iterable[builtins.str] | None = ...,
            num_buckets: builtins.int = ...,
        ) -> None: ...
        def ClearField(
            self,
            field_name: typing_extensions.Literal[
                "bucket_column_names", b"bucket_column_names", "num_buckets", b"num_buckets"
            ],
        ) -> None: ...

    INPUT_FIELD_NUMBER: builtins.int
    SOURCE_FIELD_NUMBER: builtins.int
    PATH_FIELD_NUMBER: builtins.int
    TABLE_FIELD_NUMBER: builtins.int
    MODE_FIELD_NUMBER: builtins.int
    SORT_COLUMN_NAMES_FIELD_NUMBER: builtins.int
    PARTITIONING_COLUMNS_FIELD_NUMBER: builtins.int
    BUCKET_BY_FIELD_NUMBER: builtins.int
    OPTIONS_FIELD_NUMBER: builtins.int
    CLUSTERING_COLUMNS_FIELD_NUMBER: builtins.int
    @property
    def input(self) -> pyspark.sql.connect.proto.relations_pb2.Relation:
        """(Required) The output of the `input` relation will be persisted according to the options."""
    source: builtins.str
    """(Optional) Format value according to the Spark documentation. Examples are: text, parquet, delta."""
    path: builtins.str
    @property
    def table(self) -> global___WriteOperation.SaveTable: ...
    mode: global___WriteOperation.SaveMode.ValueType
    """(Required) the save mode."""
    @property
    def sort_column_names(
        self,
    ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """(Optional) List of columns to sort the output by."""
    @property
    def partitioning_columns(
        self,
    ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """(Optional) List of columns for partitioning."""
    @property
    def bucket_by(self) -> global___WriteOperation.BucketBy:
        """(Optional) Bucketing specification. Bucketing must set the number of buckets and the columns
        to bucket by.
        """
    @property
    def options(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]:
        """(Optional) A list of configuration options."""
    @property
    def clustering_columns(
        self,
    ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """(Optional) Columns used for clustering the table."""
    def __init__(
        self,
        *,
        input: pyspark.sql.connect.proto.relations_pb2.Relation | None = ...,
        source: builtins.str | None = ...,
        path: builtins.str = ...,
        table: global___WriteOperation.SaveTable | None = ...,
        mode: global___WriteOperation.SaveMode.ValueType = ...,
        sort_column_names: collections.abc.Iterable[builtins.str] | None = ...,
        partitioning_columns: collections.abc.Iterable[builtins.str] | None = ...,
        bucket_by: global___WriteOperation.BucketBy | None = ...,
        options: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
        clustering_columns: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "_source",
            b"_source",
            "bucket_by",
            b"bucket_by",
            "input",
            b"input",
            "path",
            b"path",
            "save_type",
            b"save_type",
            "source",
            b"source",
            "table",
            b"table",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "_source",
            b"_source",
            "bucket_by",
            b"bucket_by",
            "clustering_columns",
            b"clustering_columns",
            "input",
            b"input",
            "mode",
            b"mode",
            "options",
            b"options",
            "partitioning_columns",
            b"partitioning_columns",
            "path",
            b"path",
            "save_type",
            b"save_type",
            "sort_column_names",
            b"sort_column_names",
            "source",
            b"source",
            "table",
            b"table",
        ],
    ) -> None: ...
    @typing.overload
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["_source", b"_source"]
    ) -> typing_extensions.Literal["source"] | None: ...
    @typing.overload
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["save_type", b"save_type"]
    ) -> typing_extensions.Literal["path", "table"] | None: ...

global___WriteOperation = WriteOperation

class WriteOperationV2(google.protobuf.message.Message):
    """As writes are not directly handled during analysis and planning, they are modeled as commands."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _Mode:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _ModeEnumTypeWrapper(
        google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[
            WriteOperationV2._Mode.ValueType
        ],
        builtins.type,
    ):  # noqa: F821
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        MODE_UNSPECIFIED: WriteOperationV2._Mode.ValueType  # 0
        MODE_CREATE: WriteOperationV2._Mode.ValueType  # 1
        MODE_OVERWRITE: WriteOperationV2._Mode.ValueType  # 2
        MODE_OVERWRITE_PARTITIONS: WriteOperationV2._Mode.ValueType  # 3
        MODE_APPEND: WriteOperationV2._Mode.ValueType  # 4
        MODE_REPLACE: WriteOperationV2._Mode.ValueType  # 5
        MODE_CREATE_OR_REPLACE: WriteOperationV2._Mode.ValueType  # 6

    class Mode(_Mode, metaclass=_ModeEnumTypeWrapper): ...
    MODE_UNSPECIFIED: WriteOperationV2.Mode.ValueType  # 0
    MODE_CREATE: WriteOperationV2.Mode.ValueType  # 1
    MODE_OVERWRITE: WriteOperationV2.Mode.ValueType  # 2
    MODE_OVERWRITE_PARTITIONS: WriteOperationV2.Mode.ValueType  # 3
    MODE_APPEND: WriteOperationV2.Mode.ValueType  # 4
    MODE_REPLACE: WriteOperationV2.Mode.ValueType  # 5
    MODE_CREATE_OR_REPLACE: WriteOperationV2.Mode.ValueType  # 6

    class OptionsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(
            self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]
        ) -> None: ...

    class TablePropertiesEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(
            self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]
        ) -> None: ...

    INPUT_FIELD_NUMBER: builtins.int
    TABLE_NAME_FIELD_NUMBER: builtins.int
    PROVIDER_FIELD_NUMBER: builtins.int
    PARTITIONING_COLUMNS_FIELD_NUMBER: builtins.int
    OPTIONS_FIELD_NUMBER: builtins.int
    TABLE_PROPERTIES_FIELD_NUMBER: builtins.int
    MODE_FIELD_NUMBER: builtins.int
    OVERWRITE_CONDITION_FIELD_NUMBER: builtins.int
    CLUSTERING_COLUMNS_FIELD_NUMBER: builtins.int
    @property
    def input(self) -> pyspark.sql.connect.proto.relations_pb2.Relation:
        """(Required) The output of the `input` relation will be persisted according to the options."""
    table_name: builtins.str
    """(Required) The destination of the write operation must be either a path or a table."""
    provider: builtins.str
    """(Optional) A provider for the underlying output data source. Spark's default catalog supports
    "parquet", "json", etc.
    """
    @property
    def partitioning_columns(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
        pyspark.sql.connect.proto.expressions_pb2.Expression
    ]:
        """(Optional) List of columns for partitioning for output table created by `create`,
        `createOrReplace`, or `replace`
        """
    @property
    def options(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]:
        """(Optional) A list of configuration options."""
    @property
    def table_properties(
        self,
    ) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]:
        """(Optional) A list of table properties."""
    mode: global___WriteOperationV2.Mode.ValueType
    """(Required) Write mode."""
    @property
    def overwrite_condition(self) -> pyspark.sql.connect.proto.expressions_pb2.Expression:
        """(Optional) A condition for overwrite saving mode"""
    @property
    def clustering_columns(
        self,
    ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """(Optional) Columns used for clustering the table."""
    def __init__(
        self,
        *,
        input: pyspark.sql.connect.proto.relations_pb2.Relation | None = ...,
        table_name: builtins.str = ...,
        provider: builtins.str | None = ...,
        partitioning_columns: collections.abc.Iterable[
            pyspark.sql.connect.proto.expressions_pb2.Expression
        ]
        | None = ...,
        options: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
        table_properties: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
        mode: global___WriteOperationV2.Mode.ValueType = ...,
        overwrite_condition: pyspark.sql.connect.proto.expressions_pb2.Expression | None = ...,
        clustering_columns: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "_provider",
            b"_provider",
            "input",
            b"input",
            "overwrite_condition",
            b"overwrite_condition",
            "provider",
            b"provider",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "_provider",
            b"_provider",
            "clustering_columns",
            b"clustering_columns",
            "input",
            b"input",
            "mode",
            b"mode",
            "options",
            b"options",
            "overwrite_condition",
            b"overwrite_condition",
            "partitioning_columns",
            b"partitioning_columns",
            "provider",
            b"provider",
            "table_name",
            b"table_name",
            "table_properties",
            b"table_properties",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["_provider", b"_provider"]
    ) -> typing_extensions.Literal["provider"] | None: ...

global___WriteOperationV2 = WriteOperationV2

class WriteStreamOperationStart(google.protobuf.message.Message):
    """Starts write stream operation as streaming query. Query ID and Run ID of the streaming
    query are returned.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class OptionsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(
            self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]
        ) -> None: ...

    INPUT_FIELD_NUMBER: builtins.int
    FORMAT_FIELD_NUMBER: builtins.int
    OPTIONS_FIELD_NUMBER: builtins.int
    PARTITIONING_COLUMN_NAMES_FIELD_NUMBER: builtins.int
    PROCESSING_TIME_INTERVAL_FIELD_NUMBER: builtins.int
    AVAILABLE_NOW_FIELD_NUMBER: builtins.int
    ONCE_FIELD_NUMBER: builtins.int
    CONTINUOUS_CHECKPOINT_INTERVAL_FIELD_NUMBER: builtins.int
    OUTPUT_MODE_FIELD_NUMBER: builtins.int
    QUERY_NAME_FIELD_NUMBER: builtins.int
    PATH_FIELD_NUMBER: builtins.int
    TABLE_NAME_FIELD_NUMBER: builtins.int
    FOREACH_WRITER_FIELD_NUMBER: builtins.int
    FOREACH_BATCH_FIELD_NUMBER: builtins.int
    CLUSTERING_COLUMN_NAMES_FIELD_NUMBER: builtins.int
    @property
    def input(self) -> pyspark.sql.connect.proto.relations_pb2.Relation:
        """(Required) The output of the `input` streaming relation will be written."""
    format: builtins.str
    """The following fields directly map to API for DataStreamWriter().
    Consult API documentation unless explicitly documented here.
    """
    @property
    def options(
        self,
    ) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]: ...
    @property
    def partitioning_column_names(
        self,
    ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    processing_time_interval: builtins.str
    available_now: builtins.bool
    once: builtins.bool
    continuous_checkpoint_interval: builtins.str
    output_mode: builtins.str
    query_name: builtins.str
    path: builtins.str
    table_name: builtins.str
    @property
    def foreach_writer(self) -> global___StreamingForeachFunction: ...
    @property
    def foreach_batch(self) -> global___StreamingForeachFunction: ...
    @property
    def clustering_column_names(
        self,
    ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """(Optional) Columns used for clustering the table."""
    def __init__(
        self,
        *,
        input: pyspark.sql.connect.proto.relations_pb2.Relation | None = ...,
        format: builtins.str = ...,
        options: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
        partitioning_column_names: collections.abc.Iterable[builtins.str] | None = ...,
        processing_time_interval: builtins.str = ...,
        available_now: builtins.bool = ...,
        once: builtins.bool = ...,
        continuous_checkpoint_interval: builtins.str = ...,
        output_mode: builtins.str = ...,
        query_name: builtins.str = ...,
        path: builtins.str = ...,
        table_name: builtins.str = ...,
        foreach_writer: global___StreamingForeachFunction | None = ...,
        foreach_batch: global___StreamingForeachFunction | None = ...,
        clustering_column_names: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "available_now",
            b"available_now",
            "continuous_checkpoint_interval",
            b"continuous_checkpoint_interval",
            "foreach_batch",
            b"foreach_batch",
            "foreach_writer",
            b"foreach_writer",
            "input",
            b"input",
            "once",
            b"once",
            "path",
            b"path",
            "processing_time_interval",
            b"processing_time_interval",
            "sink_destination",
            b"sink_destination",
            "table_name",
            b"table_name",
            "trigger",
            b"trigger",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "available_now",
            b"available_now",
            "clustering_column_names",
            b"clustering_column_names",
            "continuous_checkpoint_interval",
            b"continuous_checkpoint_interval",
            "foreach_batch",
            b"foreach_batch",
            "foreach_writer",
            b"foreach_writer",
            "format",
            b"format",
            "input",
            b"input",
            "once",
            b"once",
            "options",
            b"options",
            "output_mode",
            b"output_mode",
            "partitioning_column_names",
            b"partitioning_column_names",
            "path",
            b"path",
            "processing_time_interval",
            b"processing_time_interval",
            "query_name",
            b"query_name",
            "sink_destination",
            b"sink_destination",
            "table_name",
            b"table_name",
            "trigger",
            b"trigger",
        ],
    ) -> None: ...
    @typing.overload
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["sink_destination", b"sink_destination"]
    ) -> typing_extensions.Literal["path", "table_name"] | None: ...
    @typing.overload
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["trigger", b"trigger"]
    ) -> (
        typing_extensions.Literal[
            "processing_time_interval", "available_now", "once", "continuous_checkpoint_interval"
        ]
        | None
    ): ...

global___WriteStreamOperationStart = WriteStreamOperationStart

class StreamingForeachFunction(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PYTHON_FUNCTION_FIELD_NUMBER: builtins.int
    SCALA_FUNCTION_FIELD_NUMBER: builtins.int
    @property
    def python_function(self) -> pyspark.sql.connect.proto.expressions_pb2.PythonUDF: ...
    @property
    def scala_function(self) -> pyspark.sql.connect.proto.expressions_pb2.ScalarScalaUDF: ...
    def __init__(
        self,
        *,
        python_function: pyspark.sql.connect.proto.expressions_pb2.PythonUDF | None = ...,
        scala_function: pyspark.sql.connect.proto.expressions_pb2.ScalarScalaUDF | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "function",
            b"function",
            "python_function",
            b"python_function",
            "scala_function",
            b"scala_function",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "function",
            b"function",
            "python_function",
            b"python_function",
            "scala_function",
            b"scala_function",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["function", b"function"]
    ) -> typing_extensions.Literal["python_function", "scala_function"] | None: ...

global___StreamingForeachFunction = StreamingForeachFunction

class WriteStreamOperationStartResult(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    QUERY_ID_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    QUERY_STARTED_EVENT_JSON_FIELD_NUMBER: builtins.int
    @property
    def query_id(self) -> global___StreamingQueryInstanceId:
        """(Required) Query instance. See `StreamingQueryInstanceId`."""
    name: builtins.str
    """An optional query name."""
    query_started_event_json: builtins.str
    """Optional query started event if there is any listener registered on the client side."""
    def __init__(
        self,
        *,
        query_id: global___StreamingQueryInstanceId | None = ...,
        name: builtins.str = ...,
        query_started_event_json: builtins.str | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "_query_started_event_json",
            b"_query_started_event_json",
            "query_id",
            b"query_id",
            "query_started_event_json",
            b"query_started_event_json",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "_query_started_event_json",
            b"_query_started_event_json",
            "name",
            b"name",
            "query_id",
            b"query_id",
            "query_started_event_json",
            b"query_started_event_json",
        ],
    ) -> None: ...
    def WhichOneof(
        self,
        oneof_group: typing_extensions.Literal[
            "_query_started_event_json", b"_query_started_event_json"
        ],
    ) -> typing_extensions.Literal["query_started_event_json"] | None: ...

global___WriteStreamOperationStartResult = WriteStreamOperationStartResult

class StreamingQueryInstanceId(google.protobuf.message.Message):
    """A tuple that uniquely identifies an instance of streaming query run. It consists of `id` that
    persists across the streaming runs and `run_id` that changes between each run of the
    streaming query that resumes from the checkpoint.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ID_FIELD_NUMBER: builtins.int
    RUN_ID_FIELD_NUMBER: builtins.int
    id: builtins.str
    """(Required) The unique id of this query that persists across restarts from checkpoint data.
    That is, this id is generated when a query is started for the first time, and
    will be the same every time it is restarted from checkpoint data.
    """
    run_id: builtins.str
    """(Required) The unique id of this run of the query. That is, every start/restart of a query
    will generate a unique run_id. Therefore, every time a query is restarted from
    checkpoint, it will have the same `id` but different `run_id`s.
    """
    def __init__(
        self,
        *,
        id: builtins.str = ...,
        run_id: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["id", b"id", "run_id", b"run_id"]
    ) -> None: ...

global___StreamingQueryInstanceId = StreamingQueryInstanceId

class StreamingQueryCommand(google.protobuf.message.Message):
    """Commands for a streaming query."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class ExplainCommand(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        EXTENDED_FIELD_NUMBER: builtins.int
        extended: builtins.bool
        """TODO: Consider reusing Explain from AnalyzePlanRequest message.
              We can not do this right now since it base.proto imports this file.
        """
        def __init__(
            self,
            *,
            extended: builtins.bool = ...,
        ) -> None: ...
        def ClearField(
            self, field_name: typing_extensions.Literal["extended", b"extended"]
        ) -> None: ...

    class AwaitTerminationCommand(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        TIMEOUT_MS_FIELD_NUMBER: builtins.int
        timeout_ms: builtins.int
        def __init__(
            self,
            *,
            timeout_ms: builtins.int | None = ...,
        ) -> None: ...
        def HasField(
            self,
            field_name: typing_extensions.Literal[
                "_timeout_ms", b"_timeout_ms", "timeout_ms", b"timeout_ms"
            ],
        ) -> builtins.bool: ...
        def ClearField(
            self,
            field_name: typing_extensions.Literal[
                "_timeout_ms", b"_timeout_ms", "timeout_ms", b"timeout_ms"
            ],
        ) -> None: ...
        def WhichOneof(
            self, oneof_group: typing_extensions.Literal["_timeout_ms", b"_timeout_ms"]
        ) -> typing_extensions.Literal["timeout_ms"] | None: ...

    QUERY_ID_FIELD_NUMBER: builtins.int
    STATUS_FIELD_NUMBER: builtins.int
    LAST_PROGRESS_FIELD_NUMBER: builtins.int
    RECENT_PROGRESS_FIELD_NUMBER: builtins.int
    STOP_FIELD_NUMBER: builtins.int
    PROCESS_ALL_AVAILABLE_FIELD_NUMBER: builtins.int
    EXPLAIN_FIELD_NUMBER: builtins.int
    EXCEPTION_FIELD_NUMBER: builtins.int
    AWAIT_TERMINATION_FIELD_NUMBER: builtins.int
    @property
    def query_id(self) -> global___StreamingQueryInstanceId:
        """(Required) Query instance. See `StreamingQueryInstanceId`."""
    status: builtins.bool
    """status() API."""
    last_progress: builtins.bool
    """lastProgress() API."""
    recent_progress: builtins.bool
    """recentProgress() API."""
    stop: builtins.bool
    """stop() API. Stops the query."""
    process_all_available: builtins.bool
    """processAllAvailable() API. Waits till all the available data is processed"""
    @property
    def explain(self) -> global___StreamingQueryCommand.ExplainCommand:
        """explain() API. Returns logical and physical plans."""
    exception: builtins.bool
    """exception() API. Returns the exception in the query if any."""
    @property
    def await_termination(self) -> global___StreamingQueryCommand.AwaitTerminationCommand:
        """awaitTermination() API. Waits for the termination of the query."""
    def __init__(
        self,
        *,
        query_id: global___StreamingQueryInstanceId | None = ...,
        status: builtins.bool = ...,
        last_progress: builtins.bool = ...,
        recent_progress: builtins.bool = ...,
        stop: builtins.bool = ...,
        process_all_available: builtins.bool = ...,
        explain: global___StreamingQueryCommand.ExplainCommand | None = ...,
        exception: builtins.bool = ...,
        await_termination: global___StreamingQueryCommand.AwaitTerminationCommand | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "await_termination",
            b"await_termination",
            "command",
            b"command",
            "exception",
            b"exception",
            "explain",
            b"explain",
            "last_progress",
            b"last_progress",
            "process_all_available",
            b"process_all_available",
            "query_id",
            b"query_id",
            "recent_progress",
            b"recent_progress",
            "status",
            b"status",
            "stop",
            b"stop",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "await_termination",
            b"await_termination",
            "command",
            b"command",
            "exception",
            b"exception",
            "explain",
            b"explain",
            "last_progress",
            b"last_progress",
            "process_all_available",
            b"process_all_available",
            "query_id",
            b"query_id",
            "recent_progress",
            b"recent_progress",
            "status",
            b"status",
            "stop",
            b"stop",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["command", b"command"]
    ) -> (
        typing_extensions.Literal[
            "status",
            "last_progress",
            "recent_progress",
            "stop",
            "process_all_available",
            "explain",
            "exception",
            "await_termination",
        ]
        | None
    ): ...

global___StreamingQueryCommand = StreamingQueryCommand

class StreamingQueryCommandResult(google.protobuf.message.Message):
    """Response for commands on a streaming query."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class StatusResult(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        STATUS_MESSAGE_FIELD_NUMBER: builtins.int
        IS_DATA_AVAILABLE_FIELD_NUMBER: builtins.int
        IS_TRIGGER_ACTIVE_FIELD_NUMBER: builtins.int
        IS_ACTIVE_FIELD_NUMBER: builtins.int
        status_message: builtins.str
        """See documentation for these Scala 'StreamingQueryStatus' struct"""
        is_data_available: builtins.bool
        is_trigger_active: builtins.bool
        is_active: builtins.bool
        def __init__(
            self,
            *,
            status_message: builtins.str = ...,
            is_data_available: builtins.bool = ...,
            is_trigger_active: builtins.bool = ...,
            is_active: builtins.bool = ...,
        ) -> None: ...
        def ClearField(
            self,
            field_name: typing_extensions.Literal[
                "is_active",
                b"is_active",
                "is_data_available",
                b"is_data_available",
                "is_trigger_active",
                b"is_trigger_active",
                "status_message",
                b"status_message",
            ],
        ) -> None: ...

    class RecentProgressResult(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        RECENT_PROGRESS_JSON_FIELD_NUMBER: builtins.int
        @property
        def recent_progress_json(
            self,
        ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
            """Progress reports as an array of json strings."""
        def __init__(
            self,
            *,
            recent_progress_json: collections.abc.Iterable[builtins.str] | None = ...,
        ) -> None: ...
        def ClearField(
            self,
            field_name: typing_extensions.Literal["recent_progress_json", b"recent_progress_json"],
        ) -> None: ...

    class ExplainResult(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        RESULT_FIELD_NUMBER: builtins.int
        result: builtins.str
        """Logical and physical plans as string"""
        def __init__(
            self,
            *,
            result: builtins.str = ...,
        ) -> None: ...
        def ClearField(
            self, field_name: typing_extensions.Literal["result", b"result"]
        ) -> None: ...

    class ExceptionResult(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        EXCEPTION_MESSAGE_FIELD_NUMBER: builtins.int
        ERROR_CLASS_FIELD_NUMBER: builtins.int
        STACK_TRACE_FIELD_NUMBER: builtins.int
        exception_message: builtins.str
        """(Optional) Exception message as string, maps to the return value of original
        StreamingQueryException's toString method
        """
        error_class: builtins.str
        """(Optional) Exception error class as string"""
        stack_trace: builtins.str
        """(Optional) Exception stack trace as string"""
        def __init__(
            self,
            *,
            exception_message: builtins.str | None = ...,
            error_class: builtins.str | None = ...,
            stack_trace: builtins.str | None = ...,
        ) -> None: ...
        def HasField(
            self,
            field_name: typing_extensions.Literal[
                "_error_class",
                b"_error_class",
                "_exception_message",
                b"_exception_message",
                "_stack_trace",
                b"_stack_trace",
                "error_class",
                b"error_class",
                "exception_message",
                b"exception_message",
                "stack_trace",
                b"stack_trace",
            ],
        ) -> builtins.bool: ...
        def ClearField(
            self,
            field_name: typing_extensions.Literal[
                "_error_class",
                b"_error_class",
                "_exception_message",
                b"_exception_message",
                "_stack_trace",
                b"_stack_trace",
                "error_class",
                b"error_class",
                "exception_message",
                b"exception_message",
                "stack_trace",
                b"stack_trace",
            ],
        ) -> None: ...
        @typing.overload
        def WhichOneof(
            self, oneof_group: typing_extensions.Literal["_error_class", b"_error_class"]
        ) -> typing_extensions.Literal["error_class"] | None: ...
        @typing.overload
        def WhichOneof(
            self,
            oneof_group: typing_extensions.Literal["_exception_message", b"_exception_message"],
        ) -> typing_extensions.Literal["exception_message"] | None: ...
        @typing.overload
        def WhichOneof(
            self, oneof_group: typing_extensions.Literal["_stack_trace", b"_stack_trace"]
        ) -> typing_extensions.Literal["stack_trace"] | None: ...

    class AwaitTerminationResult(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        TERMINATED_FIELD_NUMBER: builtins.int
        terminated: builtins.bool
        def __init__(
            self,
            *,
            terminated: builtins.bool = ...,
        ) -> None: ...
        def ClearField(
            self, field_name: typing_extensions.Literal["terminated", b"terminated"]
        ) -> None: ...

    QUERY_ID_FIELD_NUMBER: builtins.int
    STATUS_FIELD_NUMBER: builtins.int
    RECENT_PROGRESS_FIELD_NUMBER: builtins.int
    EXPLAIN_FIELD_NUMBER: builtins.int
    EXCEPTION_FIELD_NUMBER: builtins.int
    AWAIT_TERMINATION_FIELD_NUMBER: builtins.int
    @property
    def query_id(self) -> global___StreamingQueryInstanceId:
        """(Required) Query instance id. See `StreamingQueryInstanceId`."""
    @property
    def status(self) -> global___StreamingQueryCommandResult.StatusResult: ...
    @property
    def recent_progress(self) -> global___StreamingQueryCommandResult.RecentProgressResult: ...
    @property
    def explain(self) -> global___StreamingQueryCommandResult.ExplainResult: ...
    @property
    def exception(self) -> global___StreamingQueryCommandResult.ExceptionResult: ...
    @property
    def await_termination(self) -> global___StreamingQueryCommandResult.AwaitTerminationResult: ...
    def __init__(
        self,
        *,
        query_id: global___StreamingQueryInstanceId | None = ...,
        status: global___StreamingQueryCommandResult.StatusResult | None = ...,
        recent_progress: global___StreamingQueryCommandResult.RecentProgressResult | None = ...,
        explain: global___StreamingQueryCommandResult.ExplainResult | None = ...,
        exception: global___StreamingQueryCommandResult.ExceptionResult | None = ...,
        await_termination: global___StreamingQueryCommandResult.AwaitTerminationResult | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "await_termination",
            b"await_termination",
            "exception",
            b"exception",
            "explain",
            b"explain",
            "query_id",
            b"query_id",
            "recent_progress",
            b"recent_progress",
            "result_type",
            b"result_type",
            "status",
            b"status",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "await_termination",
            b"await_termination",
            "exception",
            b"exception",
            "explain",
            b"explain",
            "query_id",
            b"query_id",
            "recent_progress",
            b"recent_progress",
            "result_type",
            b"result_type",
            "status",
            b"status",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["result_type", b"result_type"]
    ) -> (
        typing_extensions.Literal[
            "status", "recent_progress", "explain", "exception", "await_termination"
        ]
        | None
    ): ...

global___StreamingQueryCommandResult = StreamingQueryCommandResult

class StreamingQueryManagerCommand(google.protobuf.message.Message):
    """Commands for the streaming query manager."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class AwaitAnyTerminationCommand(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        TIMEOUT_MS_FIELD_NUMBER: builtins.int
        timeout_ms: builtins.int
        """(Optional) The waiting time in milliseconds to wait for any query to terminate."""
        def __init__(
            self,
            *,
            timeout_ms: builtins.int | None = ...,
        ) -> None: ...
        def HasField(
            self,
            field_name: typing_extensions.Literal[
                "_timeout_ms", b"_timeout_ms", "timeout_ms", b"timeout_ms"
            ],
        ) -> builtins.bool: ...
        def ClearField(
            self,
            field_name: typing_extensions.Literal[
                "_timeout_ms", b"_timeout_ms", "timeout_ms", b"timeout_ms"
            ],
        ) -> None: ...
        def WhichOneof(
            self, oneof_group: typing_extensions.Literal["_timeout_ms", b"_timeout_ms"]
        ) -> typing_extensions.Literal["timeout_ms"] | None: ...

    class StreamingQueryListenerCommand(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        LISTENER_PAYLOAD_FIELD_NUMBER: builtins.int
        PYTHON_LISTENER_PAYLOAD_FIELD_NUMBER: builtins.int
        ID_FIELD_NUMBER: builtins.int
        listener_payload: builtins.bytes
        @property
        def python_listener_payload(
            self,
        ) -> pyspark.sql.connect.proto.expressions_pb2.PythonUDF: ...
        id: builtins.str
        def __init__(
            self,
            *,
            listener_payload: builtins.bytes = ...,
            python_listener_payload: pyspark.sql.connect.proto.expressions_pb2.PythonUDF
            | None = ...,
            id: builtins.str = ...,
        ) -> None: ...
        def HasField(
            self,
            field_name: typing_extensions.Literal[
                "_python_listener_payload",
                b"_python_listener_payload",
                "python_listener_payload",
                b"python_listener_payload",
            ],
        ) -> builtins.bool: ...
        def ClearField(
            self,
            field_name: typing_extensions.Literal[
                "_python_listener_payload",
                b"_python_listener_payload",
                "id",
                b"id",
                "listener_payload",
                b"listener_payload",
                "python_listener_payload",
                b"python_listener_payload",
            ],
        ) -> None: ...
        def WhichOneof(
            self,
            oneof_group: typing_extensions.Literal[
                "_python_listener_payload", b"_python_listener_payload"
            ],
        ) -> typing_extensions.Literal["python_listener_payload"] | None: ...

    ACTIVE_FIELD_NUMBER: builtins.int
    GET_QUERY_FIELD_NUMBER: builtins.int
    AWAIT_ANY_TERMINATION_FIELD_NUMBER: builtins.int
    RESET_TERMINATED_FIELD_NUMBER: builtins.int
    ADD_LISTENER_FIELD_NUMBER: builtins.int
    REMOVE_LISTENER_FIELD_NUMBER: builtins.int
    LIST_LISTENERS_FIELD_NUMBER: builtins.int
    active: builtins.bool
    """active() API, returns a list of active queries."""
    get_query: builtins.str
    """get() API, returns the StreamingQuery identified by id."""
    @property
    def await_any_termination(
        self,
    ) -> global___StreamingQueryManagerCommand.AwaitAnyTerminationCommand:
        """awaitAnyTermination() API, wait until any query terminates or timeout."""
    reset_terminated: builtins.bool
    """resetTerminated() API."""
    @property
    def add_listener(self) -> global___StreamingQueryManagerCommand.StreamingQueryListenerCommand:
        """addListener API."""
    @property
    def remove_listener(
        self,
    ) -> global___StreamingQueryManagerCommand.StreamingQueryListenerCommand:
        """removeListener API."""
    list_listeners: builtins.bool
    """listListeners() API, returns a list of streaming query listeners."""
    def __init__(
        self,
        *,
        active: builtins.bool = ...,
        get_query: builtins.str = ...,
        await_any_termination: global___StreamingQueryManagerCommand.AwaitAnyTerminationCommand
        | None = ...,
        reset_terminated: builtins.bool = ...,
        add_listener: global___StreamingQueryManagerCommand.StreamingQueryListenerCommand
        | None = ...,
        remove_listener: global___StreamingQueryManagerCommand.StreamingQueryListenerCommand
        | None = ...,
        list_listeners: builtins.bool = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "active",
            b"active",
            "add_listener",
            b"add_listener",
            "await_any_termination",
            b"await_any_termination",
            "command",
            b"command",
            "get_query",
            b"get_query",
            "list_listeners",
            b"list_listeners",
            "remove_listener",
            b"remove_listener",
            "reset_terminated",
            b"reset_terminated",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "active",
            b"active",
            "add_listener",
            b"add_listener",
            "await_any_termination",
            b"await_any_termination",
            "command",
            b"command",
            "get_query",
            b"get_query",
            "list_listeners",
            b"list_listeners",
            "remove_listener",
            b"remove_listener",
            "reset_terminated",
            b"reset_terminated",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["command", b"command"]
    ) -> (
        typing_extensions.Literal[
            "active",
            "get_query",
            "await_any_termination",
            "reset_terminated",
            "add_listener",
            "remove_listener",
            "list_listeners",
        ]
        | None
    ): ...

global___StreamingQueryManagerCommand = StreamingQueryManagerCommand

class StreamingQueryManagerCommandResult(google.protobuf.message.Message):
    """Response for commands on the streaming query manager."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class ActiveResult(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        ACTIVE_QUERIES_FIELD_NUMBER: builtins.int
        @property
        def active_queries(
            self,
        ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
            global___StreamingQueryManagerCommandResult.StreamingQueryInstance
        ]: ...
        def __init__(
            self,
            *,
            active_queries: collections.abc.Iterable[
                global___StreamingQueryManagerCommandResult.StreamingQueryInstance
            ]
            | None = ...,
        ) -> None: ...
        def ClearField(
            self, field_name: typing_extensions.Literal["active_queries", b"active_queries"]
        ) -> None: ...

    class StreamingQueryInstance(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        ID_FIELD_NUMBER: builtins.int
        NAME_FIELD_NUMBER: builtins.int
        @property
        def id(self) -> global___StreamingQueryInstanceId:
            """(Required) The id and runId of this query."""
        name: builtins.str
        """(Optional) The name of this query."""
        def __init__(
            self,
            *,
            id: global___StreamingQueryInstanceId | None = ...,
            name: builtins.str | None = ...,
        ) -> None: ...
        def HasField(
            self,
            field_name: typing_extensions.Literal["_name", b"_name", "id", b"id", "name", b"name"],
        ) -> builtins.bool: ...
        def ClearField(
            self,
            field_name: typing_extensions.Literal["_name", b"_name", "id", b"id", "name", b"name"],
        ) -> None: ...
        def WhichOneof(
            self, oneof_group: typing_extensions.Literal["_name", b"_name"]
        ) -> typing_extensions.Literal["name"] | None: ...

    class AwaitAnyTerminationResult(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        TERMINATED_FIELD_NUMBER: builtins.int
        terminated: builtins.bool
        def __init__(
            self,
            *,
            terminated: builtins.bool = ...,
        ) -> None: ...
        def ClearField(
            self, field_name: typing_extensions.Literal["terminated", b"terminated"]
        ) -> None: ...

    class StreamingQueryListenerInstance(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        LISTENER_PAYLOAD_FIELD_NUMBER: builtins.int
        listener_payload: builtins.bytes
        def __init__(
            self,
            *,
            listener_payload: builtins.bytes = ...,
        ) -> None: ...
        def ClearField(
            self, field_name: typing_extensions.Literal["listener_payload", b"listener_payload"]
        ) -> None: ...

    class ListStreamingQueryListenerResult(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        LISTENER_IDS_FIELD_NUMBER: builtins.int
        @property
        def listener_ids(
            self,
        ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
            """(Required) Reference IDs of listener instances."""
        def __init__(
            self,
            *,
            listener_ids: collections.abc.Iterable[builtins.str] | None = ...,
        ) -> None: ...
        def ClearField(
            self, field_name: typing_extensions.Literal["listener_ids", b"listener_ids"]
        ) -> None: ...

    ACTIVE_FIELD_NUMBER: builtins.int
    QUERY_FIELD_NUMBER: builtins.int
    AWAIT_ANY_TERMINATION_FIELD_NUMBER: builtins.int
    RESET_TERMINATED_FIELD_NUMBER: builtins.int
    ADD_LISTENER_FIELD_NUMBER: builtins.int
    REMOVE_LISTENER_FIELD_NUMBER: builtins.int
    LIST_LISTENERS_FIELD_NUMBER: builtins.int
    @property
    def active(self) -> global___StreamingQueryManagerCommandResult.ActiveResult: ...
    @property
    def query(self) -> global___StreamingQueryManagerCommandResult.StreamingQueryInstance: ...
    @property
    def await_any_termination(
        self,
    ) -> global___StreamingQueryManagerCommandResult.AwaitAnyTerminationResult: ...
    reset_terminated: builtins.bool
    add_listener: builtins.bool
    remove_listener: builtins.bool
    @property
    def list_listeners(
        self,
    ) -> global___StreamingQueryManagerCommandResult.ListStreamingQueryListenerResult: ...
    def __init__(
        self,
        *,
        active: global___StreamingQueryManagerCommandResult.ActiveResult | None = ...,
        query: global___StreamingQueryManagerCommandResult.StreamingQueryInstance | None = ...,
        await_any_termination: global___StreamingQueryManagerCommandResult.AwaitAnyTerminationResult
        | None = ...,
        reset_terminated: builtins.bool = ...,
        add_listener: builtins.bool = ...,
        remove_listener: builtins.bool = ...,
        list_listeners: global___StreamingQueryManagerCommandResult.ListStreamingQueryListenerResult
        | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "active",
            b"active",
            "add_listener",
            b"add_listener",
            "await_any_termination",
            b"await_any_termination",
            "list_listeners",
            b"list_listeners",
            "query",
            b"query",
            "remove_listener",
            b"remove_listener",
            "reset_terminated",
            b"reset_terminated",
            "result_type",
            b"result_type",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "active",
            b"active",
            "add_listener",
            b"add_listener",
            "await_any_termination",
            b"await_any_termination",
            "list_listeners",
            b"list_listeners",
            "query",
            b"query",
            "remove_listener",
            b"remove_listener",
            "reset_terminated",
            b"reset_terminated",
            "result_type",
            b"result_type",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["result_type", b"result_type"]
    ) -> (
        typing_extensions.Literal[
            "active",
            "query",
            "await_any_termination",
            "reset_terminated",
            "add_listener",
            "remove_listener",
            "list_listeners",
        ]
        | None
    ): ...

global___StreamingQueryManagerCommandResult = StreamingQueryManagerCommandResult

class StreamingQueryListenerBusCommand(google.protobuf.message.Message):
    """The protocol for client-side StreamingQueryListener.
    This command will only be set when either the first listener is added to the client, or the last
    listener is removed from the client.
    The add_listener_bus_listener command will only be set true in the first case.
    The remove_listener_bus_listener command will only be set true in the second case.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ADD_LISTENER_BUS_LISTENER_FIELD_NUMBER: builtins.int
    REMOVE_LISTENER_BUS_LISTENER_FIELD_NUMBER: builtins.int
    add_listener_bus_listener: builtins.bool
    remove_listener_bus_listener: builtins.bool
    def __init__(
        self,
        *,
        add_listener_bus_listener: builtins.bool = ...,
        remove_listener_bus_listener: builtins.bool = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "add_listener_bus_listener",
            b"add_listener_bus_listener",
            "command",
            b"command",
            "remove_listener_bus_listener",
            b"remove_listener_bus_listener",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "add_listener_bus_listener",
            b"add_listener_bus_listener",
            "command",
            b"command",
            "remove_listener_bus_listener",
            b"remove_listener_bus_listener",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["command", b"command"]
    ) -> (
        typing_extensions.Literal["add_listener_bus_listener", "remove_listener_bus_listener"]
        | None
    ): ...

global___StreamingQueryListenerBusCommand = StreamingQueryListenerBusCommand

class StreamingQueryListenerEvent(google.protobuf.message.Message):
    """The protocol for the returned events in the long-running response channel."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    EVENT_JSON_FIELD_NUMBER: builtins.int
    EVENT_TYPE_FIELD_NUMBER: builtins.int
    event_json: builtins.str
    """(Required) The json serialized event, all StreamingQueryListener events have a json method"""
    event_type: global___StreamingQueryEventType.ValueType
    """(Required) Query event type used by client to decide how to deserialize the event_json"""
    def __init__(
        self,
        *,
        event_json: builtins.str = ...,
        event_type: global___StreamingQueryEventType.ValueType = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "event_json", b"event_json", "event_type", b"event_type"
        ],
    ) -> None: ...

global___StreamingQueryListenerEvent = StreamingQueryListenerEvent

class StreamingQueryListenerEventsResult(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    EVENTS_FIELD_NUMBER: builtins.int
    LISTENER_BUS_LISTENER_ADDED_FIELD_NUMBER: builtins.int
    @property
    def events(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
        global___StreamingQueryListenerEvent
    ]: ...
    listener_bus_listener_added: builtins.bool
    def __init__(
        self,
        *,
        events: collections.abc.Iterable[global___StreamingQueryListenerEvent] | None = ...,
        listener_bus_listener_added: builtins.bool | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "_listener_bus_listener_added",
            b"_listener_bus_listener_added",
            "listener_bus_listener_added",
            b"listener_bus_listener_added",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "_listener_bus_listener_added",
            b"_listener_bus_listener_added",
            "events",
            b"events",
            "listener_bus_listener_added",
            b"listener_bus_listener_added",
        ],
    ) -> None: ...
    def WhichOneof(
        self,
        oneof_group: typing_extensions.Literal[
            "_listener_bus_listener_added", b"_listener_bus_listener_added"
        ],
    ) -> typing_extensions.Literal["listener_bus_listener_added"] | None: ...

global___StreamingQueryListenerEventsResult = StreamingQueryListenerEventsResult

class GetResourcesCommand(google.protobuf.message.Message):
    """Command to get the output of 'SparkContext.resources'"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___GetResourcesCommand = GetResourcesCommand

class GetResourcesCommandResult(google.protobuf.message.Message):
    """Response for command 'GetResourcesCommand'."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class ResourcesEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> pyspark.sql.connect.proto.common_pb2.ResourceInformation: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: pyspark.sql.connect.proto.common_pb2.ResourceInformation | None = ...,
        ) -> None: ...
        def HasField(
            self, field_name: typing_extensions.Literal["value", b"value"]
        ) -> builtins.bool: ...
        def ClearField(
            self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]
        ) -> None: ...

    RESOURCES_FIELD_NUMBER: builtins.int
    @property
    def resources(
        self,
    ) -> google.protobuf.internal.containers.MessageMap[
        builtins.str, pyspark.sql.connect.proto.common_pb2.ResourceInformation
    ]: ...
    def __init__(
        self,
        *,
        resources: collections.abc.Mapping[
            builtins.str, pyspark.sql.connect.proto.common_pb2.ResourceInformation
        ]
        | None = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["resources", b"resources"]
    ) -> None: ...

global___GetResourcesCommandResult = GetResourcesCommandResult

class CreateResourceProfileCommand(google.protobuf.message.Message):
    """Command to create ResourceProfile"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PROFILE_FIELD_NUMBER: builtins.int
    @property
    def profile(self) -> pyspark.sql.connect.proto.common_pb2.ResourceProfile:
        """(Required) The ResourceProfile to be built on the server-side."""
    def __init__(
        self,
        *,
        profile: pyspark.sql.connect.proto.common_pb2.ResourceProfile | None = ...,
    ) -> None: ...
    def HasField(
        self, field_name: typing_extensions.Literal["profile", b"profile"]
    ) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["profile", b"profile"]) -> None: ...

global___CreateResourceProfileCommand = CreateResourceProfileCommand

class CreateResourceProfileCommandResult(google.protobuf.message.Message):
    """Response for command 'CreateResourceProfileCommand'."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PROFILE_ID_FIELD_NUMBER: builtins.int
    profile_id: builtins.int
    """(Required) Server-side generated resource profile id."""
    def __init__(
        self,
        *,
        profile_id: builtins.int = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["profile_id", b"profile_id"]
    ) -> None: ...

global___CreateResourceProfileCommandResult = CreateResourceProfileCommandResult

class RemoveCachedRemoteRelationCommand(google.protobuf.message.Message):
    """Command to remove `CashedRemoteRelation`"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    RELATION_FIELD_NUMBER: builtins.int
    @property
    def relation(self) -> pyspark.sql.connect.proto.relations_pb2.CachedRemoteRelation:
        """(Required) The remote to be related"""
    def __init__(
        self,
        *,
        relation: pyspark.sql.connect.proto.relations_pb2.CachedRemoteRelation | None = ...,
    ) -> None: ...
    def HasField(
        self, field_name: typing_extensions.Literal["relation", b"relation"]
    ) -> builtins.bool: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["relation", b"relation"]
    ) -> None: ...

global___RemoveCachedRemoteRelationCommand = RemoveCachedRemoteRelationCommand

class CheckpointCommand(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    RELATION_FIELD_NUMBER: builtins.int
    LOCAL_FIELD_NUMBER: builtins.int
    EAGER_FIELD_NUMBER: builtins.int
    STORAGE_LEVEL_FIELD_NUMBER: builtins.int
    @property
    def relation(self) -> pyspark.sql.connect.proto.relations_pb2.Relation:
        """(Required) The logical plan to checkpoint."""
    local: builtins.bool
    """(Required) Locally checkpoint using a local temporary
    directory in Spark Connect server (Spark Driver)
    """
    eager: builtins.bool
    """(Required) Whether to checkpoint this dataframe immediately."""
    @property
    def storage_level(self) -> pyspark.sql.connect.proto.common_pb2.StorageLevel:
        """(Optional) For local checkpoint, the storage level to use."""
    def __init__(
        self,
        *,
        relation: pyspark.sql.connect.proto.relations_pb2.Relation | None = ...,
        local: builtins.bool = ...,
        eager: builtins.bool = ...,
        storage_level: pyspark.sql.connect.proto.common_pb2.StorageLevel | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "_storage_level",
            b"_storage_level",
            "relation",
            b"relation",
            "storage_level",
            b"storage_level",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "_storage_level",
            b"_storage_level",
            "eager",
            b"eager",
            "local",
            b"local",
            "relation",
            b"relation",
            "storage_level",
            b"storage_level",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["_storage_level", b"_storage_level"]
    ) -> typing_extensions.Literal["storage_level"] | None: ...

global___CheckpointCommand = CheckpointCommand

class MergeIntoTableCommand(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TARGET_TABLE_NAME_FIELD_NUMBER: builtins.int
    SOURCE_TABLE_PLAN_FIELD_NUMBER: builtins.int
    MERGE_CONDITION_FIELD_NUMBER: builtins.int
    MATCH_ACTIONS_FIELD_NUMBER: builtins.int
    NOT_MATCHED_ACTIONS_FIELD_NUMBER: builtins.int
    NOT_MATCHED_BY_SOURCE_ACTIONS_FIELD_NUMBER: builtins.int
    WITH_SCHEMA_EVOLUTION_FIELD_NUMBER: builtins.int
    target_table_name: builtins.str
    """(Required) The name of the target table."""
    @property
    def source_table_plan(self) -> pyspark.sql.connect.proto.relations_pb2.Relation:
        """(Required) The relation of the source table."""
    @property
    def merge_condition(self) -> pyspark.sql.connect.proto.expressions_pb2.Expression:
        """(Required) The condition to match the source and target."""
    @property
    def match_actions(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
        pyspark.sql.connect.proto.expressions_pb2.Expression
    ]:
        """(Optional) The actions to be taken when the condition is matched."""
    @property
    def not_matched_actions(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
        pyspark.sql.connect.proto.expressions_pb2.Expression
    ]:
        """(Optional) The actions to be taken when the condition is not matched."""
    @property
    def not_matched_by_source_actions(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
        pyspark.sql.connect.proto.expressions_pb2.Expression
    ]:
        """(Optional) The actions to be taken when the condition is not matched by source."""
    with_schema_evolution: builtins.bool
    """(Required) Whether to enable schema evolution."""
    def __init__(
        self,
        *,
        target_table_name: builtins.str = ...,
        source_table_plan: pyspark.sql.connect.proto.relations_pb2.Relation | None = ...,
        merge_condition: pyspark.sql.connect.proto.expressions_pb2.Expression | None = ...,
        match_actions: collections.abc.Iterable[
            pyspark.sql.connect.proto.expressions_pb2.Expression
        ]
        | None = ...,
        not_matched_actions: collections.abc.Iterable[
            pyspark.sql.connect.proto.expressions_pb2.Expression
        ]
        | None = ...,
        not_matched_by_source_actions: collections.abc.Iterable[
            pyspark.sql.connect.proto.expressions_pb2.Expression
        ]
        | None = ...,
        with_schema_evolution: builtins.bool = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "merge_condition", b"merge_condition", "source_table_plan", b"source_table_plan"
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "match_actions",
            b"match_actions",
            "merge_condition",
            b"merge_condition",
            "not_matched_actions",
            b"not_matched_actions",
            "not_matched_by_source_actions",
            b"not_matched_by_source_actions",
            "source_table_plan",
            b"source_table_plan",
            "target_table_name",
            b"target_table_name",
            "with_schema_evolution",
            b"with_schema_evolution",
        ],
    ) -> None: ...

global___MergeIntoTableCommand = MergeIntoTableCommand

class ExecuteExternalCommand(google.protobuf.message.Message):
    """Execute an arbitrary string command inside an external execution engine"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class OptionsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(
            self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]
        ) -> None: ...

    RUNNER_FIELD_NUMBER: builtins.int
    COMMAND_FIELD_NUMBER: builtins.int
    OPTIONS_FIELD_NUMBER: builtins.int
    runner: builtins.str
    """(Required) The class name of the runner that implements `ExternalCommandRunner`"""
    command: builtins.str
    """(Required) The target command to be executed."""
    @property
    def options(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]:
        """(Optional) The options for the runner."""
    def __init__(
        self,
        *,
        runner: builtins.str = ...,
        command: builtins.str = ...,
        options: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "command", b"command", "options", b"options", "runner", b"runner"
        ],
    ) -> None: ...

global___ExecuteExternalCommand = ExecuteExternalCommand



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
from typing import (
    Any,
    Callable,
    List,
    Optional,
    Union,
    no_type_check,
    overload,
    TYPE_CHECKING,
)
from warnings import warn

from pyspark.errors.exceptions.captured import unwrap_spark_exception
from pyspark.util import _load_from_socket
from pyspark.sql.pandas.serializers import ArrowCollectSerializer
from pyspark.sql.pandas.types import _dedup_names
from pyspark.sql.types import (
    ArrayType,
    MapType,
    TimestampType,
    StructType,
    DataType,
    _create_row,
    StringType,
)
from pyspark.sql.utils import is_timestamp_ntz_preferred
from pyspark.traceback_utils import SCCallSiteSync
from pyspark.errors import PySparkTypeError, PySparkValueError

if TYPE_CHECKING:
    import numpy as np
    import pyarrow as pa
    from py4j.java_gateway import JavaObject

    from pyspark.sql.pandas._typing import DataFrameLike as PandasDataFrameLike
    from pyspark.sql import DataFrame


class PandasConversionMixin:
    """
    Mix-in for the conversion from Spark to pandas and PyArrow. Currently, only
    :class:`DataFrame` can use this class.
    """

    def toPandas(self) -> "PandasDataFrameLike":
        from pyspark.sql.dataframe import DataFrame

        assert isinstance(self, DataFrame)

        from pyspark.sql.pandas.types import _create_converter_to_pandas
        from pyspark.sql.pandas.utils import require_minimum_pandas_version

        require_minimum_pandas_version()

        import pandas as pd

        jconf = self.sparkSession._jconf

        if jconf.arrowPySparkEnabled():
            use_arrow = True
            try:
                from pyspark.sql.pandas.types import to_arrow_schema
                from pyspark.sql.pandas.utils import require_minimum_pyarrow_version

                require_minimum_pyarrow_version()
                to_arrow_schema(self.schema, prefers_large_types=jconf.arrowUseLargeVarTypes())
            except Exception as e:
                if jconf.arrowPySparkFallbackEnabled():
                    msg = (
                        "toPandas attempted Arrow optimization because "
                        "'spark.sql.execution.arrow.pyspark.enabled' is set to true; however, "
                        "failed by the reason below:\n  %s\n"
                        "Attempting non-optimization as "
                        "'spark.sql.execution.arrow.pyspark.fallback.enabled' is set to "
                        "true." % str(e)
                    )
                    warn(msg)
                    use_arrow = False
                else:
                    msg = (
                        "toPandas attempted Arrow optimization because "
                        "'spark.sql.execution.arrow.pyspark.enabled' is set to true, but has "
                        "reached the error below and will not continue because automatic fallback "
                        "with 'spark.sql.execution.arrow.pyspark.fallback.enabled' has been set to "
                        "false.\n  %s" % str(e)
                    )
                    warn(msg)
                    raise

            # Try to use Arrow optimization when the schema is supported and the required version
            # of PyArrow is found, if 'spark.sql.execution.arrow.pyspark.enabled' is enabled.
            if use_arrow:
                try:
                    import pyarrow as pa

                    self_destruct = jconf.arrowPySparkSelfDestructEnabled()
                    batches = self._collect_as_arrow(split_batches=self_destruct)
                    if len(batches) > 0:
                        table = pa.Table.from_batches(batches)
                        # Ensure only the table has a reference to the batches, so that
                        # self_destruct (if enabled) is effective
                        del batches
                        # Pandas DataFrame created from PyArrow uses datetime64[ns] for date type
                        # values, but we should use datetime.date to match the behavior with when
                        # Arrow optimization is disabled.
                        pandas_options = {
                            "date_as_object": True,
                            "coerce_temporal_nanoseconds": True,
                        }
                        if self_destruct:
                            # Configure PyArrow to use as little memory as possible:
                            # self_destruct - free columns as they are converted
                            # split_blocks - create a separate Pandas block for each column
                            # use_threads - convert one column at a time
                            pandas_options.update(
                                {
                                    "self_destruct": True,
                                    "split_blocks": True,
                                    "use_threads": False,
                                }
                            )
                        # Rename columns to avoid duplicated column names.
                        pdf = table.rename_columns(
                            [f"col_{i}" for i in range(table.num_columns)]
                        ).to_pandas(**pandas_options)

                        # Rename back to the original column names.
                        pdf.columns = self.columns
                    else:
                        pdf = pd.DataFrame(columns=self.columns)

                    if len(pdf.columns) > 0:
                        timezone = jconf.sessionLocalTimeZone()
                        struct_in_pandas = jconf.pandasStructHandlingMode()

                        error_on_duplicated_field_names = False
                        if struct_in_pandas == "legacy":
                            error_on_duplicated_field_names = True
                            struct_in_pandas = "dict"

                        return pd.concat(
                            [
                                _create_converter_to_pandas(
                                    field.dataType,
                                    field.nullable,
                                    timezone=timezone,
                                    struct_in_pandas=struct_in_pandas,
                                    error_on_duplicated_field_names=error_on_duplicated_field_names,
                                )(pser)
                                for (_, pser), field in zip(pdf.items(), self.schema.fields)
                            ],
                            axis="columns",
                        )
                    else:
                        return pdf
                except Exception as e:
                    # We might have to allow fallback here as well but multiple Spark jobs can
                    # be executed. So, simply fail in this case for now.
                    msg = (
                        "toPandas attempted Arrow optimization because "
                        "'spark.sql.execution.arrow.pyspark.enabled' is set to true, but has "
                        "reached the error below and can not continue. Note that "
                        "'spark.sql.execution.arrow.pyspark.fallback.enabled' does not have an "
                        "effect on failures in the middle of "
                        "computation.\n  %s" % str(e)
                    )
                    warn(msg)
                    raise

        # Below is toPandas without Arrow optimization.
        rows = self.collect()
        if len(rows) > 0:
            pdf = pd.DataFrame.from_records(
                rows, index=range(len(rows)), columns=self.columns  # type: ignore[arg-type]
            )
        else:
            pdf = pd.DataFrame(columns=self.columns)

        if len(pdf.columns) > 0:
            timezone = jconf.sessionLocalTimeZone()
            struct_in_pandas = jconf.pandasStructHandlingMode()

            return pd.concat(
                [
                    _create_converter_to_pandas(
                        field.dataType,
                        field.nullable,
                        timezone=timezone,
                        struct_in_pandas=(
                            "row" if struct_in_pandas == "legacy" else struct_in_pandas
                        ),
                        error_on_duplicated_field_names=False,
                        timestamp_utc_localized=False,
                    )(pser)
                    for (_, pser), field in zip(pdf.items(), self.schema.fields)
                ],
                axis="columns",
            )
        else:
            return pdf

    def toArrow(self) -> "pa.Table":
        from pyspark.sql.dataframe import DataFrame

        assert isinstance(self, DataFrame)

        jconf = self.sparkSession._jconf

        from pyspark.sql.pandas.types import to_arrow_schema
        from pyspark.sql.pandas.utils import require_minimum_pyarrow_version

        require_minimum_pyarrow_version()
        prefers_large_var_types = jconf.arrowUseLargeVarTypes()
        schema = to_arrow_schema(
            self.schema,
            error_on_duplicated_field_names_in_struct=True,
            prefers_large_types=prefers_large_var_types,
        )

        import pyarrow as pa

        self_destruct = jconf.arrowPySparkSelfDestructEnabled()
        batches = self._collect_as_arrow(
            split_batches=self_destruct, empty_list_if_zero_records=False
        )
        table = pa.Table.from_batches(batches).cast(schema)
        # Ensure only the table has a reference to the batches, so that
        # self_destruct (if enabled) is effective
        del batches
        return table

    def _collect_as_arrow(
        self,
        split_batches: bool = False,
        empty_list_if_zero_records: bool = True,
    ) -> List["pa.RecordBatch"]:
        """
        Returns all records as a list of Arrow RecordBatches. PyArrow must be installed
        and available on driver and worker Python environments.
        This is an experimental feature.

        :param split_batches: split batches such that each column is in its own allocation, so
            that the selfDestruct optimization is effective; default False.

        :param empty_list_if_zero_records: If True (the default), returns an empty list if the
            result has 0 records. Otherwise, returns a list of length 1 containing an empty
            Arrow RecordBatch which includes the schema.

        .. note:: Experimental.
        """
        from pyspark.sql.dataframe import DataFrame

        assert isinstance(self, DataFrame)

        with SCCallSiteSync(self._sc):
            (
                port,
                auth_secret,
                jsocket_auth_server,
            ) = self._jdf.collectAsArrowToPython()

        # Collect list of un-ordered batches where last element is a list of correct order indices
        try:
            batch_stream = _load_from_socket((port, auth_secret), ArrowCollectSerializer())
            if split_batches:
                # When spark.sql.execution.arrow.pyspark.selfDestruct.enabled, ensure
                # each column in each record batch is contained in its own allocation.
                # Otherwise, selfDestruct does nothing; it frees each column as its
                # converted, but each column will actually be a list of slices of record
                # batches, and so no memory is actually freed until all columns are
                # converted.
                import pyarrow as pa

                results = []
                for batch_or_indices in batch_stream:
                    if isinstance(batch_or_indices, pa.RecordBatch):
                        batch_or_indices = pa.RecordBatch.from_arrays(
                            [
                                # This call actually reallocates the array
                                pa.concat_arrays([array])
                                for array in batch_or_indices
                            ],
                            schema=batch_or_indices.schema,
                        )
                    results.append(batch_or_indices)
            else:
                results = list(batch_stream)
        finally:
            with unwrap_spark_exception():
                # Join serving thread and raise any exceptions from collectAsArrowToPython
                jsocket_auth_server.getResult()

        # Separate RecordBatches from batch order indices in results
        batches = results[:-1]
        batch_order = results[-1]

        if len(batches) or empty_list_if_zero_records:
            # Re-order the batch list using the correct order
            return [batches[i] for i in batch_order]
        else:
            from pyspark.sql.pandas.types import to_arrow_schema
            import pyarrow as pa

            prefers_large_var_types = self.sparkSession._jconf.arrowUseLargeVarTypes()
            schema = to_arrow_schema(self.schema, prefers_large_types=prefers_large_var_types)
            empty_arrays = [pa.array([], type=field.type) for field in schema]
            return [pa.RecordBatch.from_arrays(empty_arrays, schema=schema)]


class SparkConversionMixin:
    """
    Min-in for the conversion from pandas and PyArrow to Spark. Currently, only
    :class:`SparkSession` can use this class.
    """

    _jsparkSession: "JavaObject"

    @overload
    def createDataFrame(
        self, data: "PandasDataFrameLike", samplingRatio: Optional[float] = ...
    ) -> "DataFrame":
        ...

    @overload
    def createDataFrame(
        self, data: "pa.Table", samplingRatio: Optional[float] = ...
    ) -> "DataFrame":
        ...

    @overload
    def createDataFrame(
        self,
        data: "PandasDataFrameLike",
        schema: Union[StructType, str],
        verifySchema: bool = ...,
    ) -> "DataFrame":
        ...

    @overload
    def createDataFrame(
        self,
        data: "pa.Table",
        schema: Union[StructType, str],
        verifySchema: bool = ...,
    ) -> "DataFrame":
        ...

    def createDataFrame(  # type: ignore[misc]
        self,
        data: Union["PandasDataFrameLike", "pa.Table"],
        schema: Optional[Union[StructType, List[str]]] = None,
        samplingRatio: Optional[float] = None,
        verifySchema: bool = True,
    ) -> "DataFrame":
        from pyspark.sql import SparkSession

        assert isinstance(self, SparkSession)

        timezone = self._jconf.sessionLocalTimeZone()

        if type(data).__name__ == "Table":
            # `data` is a PyArrow Table
            from pyspark.sql.pandas.utils import require_minimum_pyarrow_version

            require_minimum_pyarrow_version()

            import pyarrow as pa

            assert isinstance(data, pa.Table)

            # If no schema supplied by user then get the names of columns only
            if schema is None:
                schema = data.schema.names

            return self._create_from_arrow_table(data, schema, timezone)

        # `data` is a PandasDataFrameLike object
        from pyspark.sql.pandas.utils import require_minimum_pandas_version

        require_minimum_pandas_version()

        # If no schema supplied by user then get the names of columns only
        if schema is None:
            schema = [str(x) if not isinstance(x, str) else x for x in data.columns]

        if self._jconf.arrowPySparkEnabled() and len(data) > 0:
            try:
                return self._create_from_pandas_with_arrow(data, schema, timezone)
            except Exception as e:
                if self._jconf.arrowPySparkFallbackEnabled():
                    msg = (
                        "createDataFrame attempted Arrow optimization because "
                        "'spark.sql.execution.arrow.pyspark.enabled' is set to true; however, "
                        "failed by the reason below:\n  %s\n"
                        "Attempting non-optimization as "
                        "'spark.sql.execution.arrow.pyspark.fallback.enabled' is set to "
                        "true." % str(e)
                    )
                    warn(msg)
                else:
                    msg = (
                        "createDataFrame attempted Arrow optimization because "
                        "'spark.sql.execution.arrow.pyspark.enabled' is set to true, but has "
                        "reached the error below and will not continue because automatic "
                        "fallback with 'spark.sql.execution.arrow.pyspark.fallback.enabled' "
                        "has been set to false.\n  %s" % str(e)
                    )
                    warn(msg)
                    raise
        converted_data = self._convert_from_pandas(data, schema, timezone)
        return self._create_dataframe(converted_data, schema, samplingRatio, verifySchema)

    def _convert_from_pandas(
        self, pdf: "PandasDataFrameLike", schema: Union[StructType, str, List[str]], timezone: str
    ) -> List:
        """
        Convert a pandas.DataFrame to list of records that can be used to make a DataFrame

        Returns
        -------
        list
            list of records
        """
        from pyspark.sql import SparkSession

        assert isinstance(self, SparkSession)

        if timezone is not None:
            from pyspark.sql.pandas.types import (
                _check_series_convert_timestamps_tz_local,
                _get_local_timezone,
            )
            import pandas as pd
            from pandas.core.dtypes.common import is_timedelta64_dtype

            copied = False
            if isinstance(schema, StructType):

                def _create_converter(data_type: DataType) -> Callable[[pd.Series], pd.Series]:
                    if isinstance(data_type, TimestampType):

                        def correct_timestamp(pser: pd.Series) -> pd.Series:
                            return _check_series_convert_timestamps_tz_local(pser, timezone)

                        return correct_timestamp

                    def _converter(dt: DataType) -> Optional[Callable[[Any], Any]]:
                        if isinstance(dt, ArrayType):
                            element_conv = _converter(dt.elementType) or (lambda x: x)

                            def convert_array(value: Any) -> Any:
                                if value is None:
                                    return None
                                else:
                                    return [element_conv(v) for v in value]

                            return convert_array

                        elif isinstance(dt, MapType):
                            key_conv = _converter(dt.keyType) or (lambda x: x)
                            value_conv = _converter(dt.valueType) or (lambda x: x)

                            def convert_map(value: Any) -> Any:
                                if value is None:
                                    return None
                                else:
                                    return {key_conv(k): value_conv(v) for k, v in value.items()}

                            return convert_map

                        elif isinstance(dt, StructType):
                            field_names = dt.names
                            dedup_field_names = _dedup_names(field_names)
                            field_convs = [
                                _converter(f.dataType) or (lambda x: x) for f in dt.fields
                            ]

                            def convert_struct(value: Any) -> Any:
                                if value is None:
                                    return None
                                elif isinstance(value, dict):
                                    _values = [
                                        field_convs[i](value.get(name, None))
                                        for i, name in enumerate(dedup_field_names)
                                    ]
                                    return _create_row(field_names, _values)
                                else:
                                    _values = [
                                        field_convs[i](value[i]) for i, name in enumerate(value)
                                    ]
                                    return _create_row(field_names, _values)

                            return convert_struct

                        elif isinstance(dt, TimestampType):

                            def convert_timestamp(value: Any) -> Any:
                                if value is None:
                                    return None
                                else:
                                    return (
                                        pd.Timestamp(value)
                                        .tz_localize(timezone, ambiguous=False)
                                        .tz_convert(_get_local_timezone())
                                        .tz_localize(None)
                                        .to_pydatetime()
                                    )

                            return convert_timestamp

                        else:
                            return None

                    conv = _converter(data_type)
                    if conv is not None:
                        return lambda pser: pser.apply(conv)  # type: ignore[return-value]
                    else:
                        return lambda pser: pser

                if len(pdf.columns) > 0:
                    pdf = pd.concat(
                        [
                            _create_converter(field.dataType)(pser)
                            for (_, pser), field in zip(pdf.items(), schema.fields)
                        ],
                        axis="columns",
                    )
                    copied = True
            else:
                should_localize = not is_timestamp_ntz_preferred()
                for column, series in pdf.items():
                    s = series
                    if (
                        should_localize
                        and isinstance(s.dtype, pd.DatetimeTZDtype)
                        and s.dt.tz is not None
                    ):
                        s = _check_series_convert_timestamps_tz_local(series, timezone)
                    if s is not series:
                        if not copied:
                            # Copy once if the series is modified to prevent the original
                            # Pandas DataFrame from being updated
                            pdf = pdf.copy()
                            copied = True
                        pdf[column] = s

            for column, series in pdf.items():
                if is_timedelta64_dtype(series):
                    if not copied:
                        pdf = pdf.copy()
                        copied = True
                    # Explicitly set the timedelta as object so the output of numpy records can
                    # hold the timedelta instances as are. Otherwise, it converts to the internal
                    # numeric values.
                    ser = pdf[column]
                    pdf[column] = pd.Series(
                        ser.dt.to_pytimedelta(), index=ser.index, dtype="object", name=ser.name
                    )

        # Convert pandas.DataFrame to list of numpy records
        np_records = pdf.set_axis(
            [f"col_{i}" for i in range(len(pdf.columns))], axis="columns"  # type: ignore[arg-type]
        ).to_records(index=False)

        # Check if any columns need to be fixed for Spark to infer properly
        if len(np_records) > 0:
            record_dtype = self._get_numpy_record_dtype(np_records[0])
            if record_dtype is not None:
                return [r.astype(record_dtype).tolist() for r in np_records]

        # Convert list of numpy records to python lists
        return [r.tolist() for r in np_records]

    def _get_numpy_record_dtype(self, rec: "np.recarray") -> Optional["np.dtype"]:
        """
        Used when converting a pandas.DataFrame to Spark using to_records(), this will correct
        the dtypes of fields in a record so they can be properly loaded into Spark.

        Parameters
        ----------
        rec : numpy.record
            a numpy record to check field dtypes

        Returns
        -------
        numpy.dtype
            corrected dtype for a numpy.record or None if no correction needed
        """
        import numpy as np

        cur_dtypes = rec.dtype
        col_names = cur_dtypes.names
        record_type_list = []
        has_rec_fix = False
        for i in range(len(cur_dtypes)):
            curr_type = cur_dtypes[i]
            # If type is a datetime64 timestamp, convert to microseconds
            # NOTE: if dtype is datetime[ns] then np.record.tolist() will output values as longs,
            # conversion from [us] or lower will lead to py datetime objects, see SPARK-22417
            if curr_type == np.dtype("datetime64[ns]"):
                curr_type = "datetime64[us]"
                has_rec_fix = True
            record_type_list.append((str(col_names[i]), curr_type))
        return np.dtype(record_type_list) if has_rec_fix else None

    def _create_from_pandas_with_arrow(
        self, pdf: "PandasDataFrameLike", schema: Union[StructType, List[str]], timezone: str
    ) -> "DataFrame":
        """
        Create a DataFrame from a given pandas.DataFrame by slicing it into partitions, converting
        to Arrow data, then sending to the JVM to parallelize. If a schema is passed in, the
        data types will be used to coerce the data in Pandas to Arrow conversion.
        """
        from pyspark.sql import SparkSession
        from pyspark.sql.dataframe import DataFrame

        assert isinstance(self, SparkSession)

        from pyspark.sql.pandas.serializers import ArrowStreamPandasSerializer
        from pyspark.sql.types import TimestampType
        from pyspark.sql.pandas.types import (
            from_arrow_type,
            to_arrow_type,
            _deduplicate_field_names,
        )
        from pyspark.sql.pandas.utils import (
            require_minimum_pandas_version,
            require_minimum_pyarrow_version,
        )

        require_minimum_pandas_version()
        require_minimum_pyarrow_version()

        import pandas as pd
        from pandas.api.types import (  # type: ignore[attr-defined]
            is_datetime64_dtype,
        )
        import pyarrow as pa

        infer_pandas_dict_as_map = (
            str(self.conf.get("spark.sql.execution.pandas.inferPandasDictAsMap")).lower() == "true"
        )

        # Create the Spark schema from list of names passed in with Arrow types
        if isinstance(schema, (list, tuple)):
            arrow_schema = pa.Schema.from_pandas(pdf, preserve_index=False)
            prefer_timestamp_ntz = is_timestamp_ntz_preferred()
            struct = StructType()
            if infer_pandas_dict_as_map:
                spark_type: Union[MapType, DataType]
                for name, field in zip(schema, arrow_schema):
                    field_type = field.type
                    if isinstance(field_type, pa.StructType):
                        if len(field_type) == 0:
                            raise PySparkValueError(
                                errorClass="CANNOT_INFER_EMPTY_SCHEMA",
                                messageParameters={},
                            )
                        arrow_type = field_type.field(0).type
                        spark_type = MapType(
                            StringType(), from_arrow_type(arrow_type, prefer_timestamp_ntz)
                        )
                    else:
                        spark_type = from_arrow_type(field_type)
                    struct.add(name, spark_type, nullable=field.nullable)
            else:
                for name, field in zip(schema, arrow_schema):
                    struct.add(
                        name,
                        from_arrow_type(field.type, prefer_timestamp_ntz),
                        nullable=field.nullable,
                    )
            schema = struct

        # Determine arrow types to coerce data when creating batches
        if isinstance(schema, StructType):
            spark_types = [_deduplicate_field_names(f.dataType) for f in schema.fields]
        elif isinstance(schema, DataType):
            raise PySparkTypeError(
                errorClass="UNSUPPORTED_DATA_TYPE_FOR_ARROW",
                messageParameters={"data_type": str(schema)},
            )
        else:
            # Any timestamps must be coerced to be compatible with Spark
            spark_types = [
                TimestampType()
                if is_datetime64_dtype(t) or isinstance(t, pd.DatetimeTZDtype)
                else None
                for t in pdf.dtypes
            ]

        # Slice the DataFrame to be batched
        step = self._jconf.arrowMaxRecordsPerBatch()
        step = step if step > 0 else len(pdf)
        pdf_slices = (pdf.iloc[start : start + step] for start in range(0, len(pdf), step))

        # Create list of Arrow (columns, arrow_type, spark_type) for serializer dump_stream
        prefers_large_var_types = self._jconf.arrowUseLargeVarTypes()
        arrow_data = [
            [
                (
                    c,
                    to_arrow_type(t, prefers_large_types=prefers_large_var_types)
                    if t is not None
                    else None,
                    t,
                )
                for (_, c), t in zip(pdf_slice.items(), spark_types)
            ]
            for pdf_slice in pdf_slices
        ]

        jsparkSession = self._jsparkSession

        safecheck = self._jconf.arrowSafeTypeConversion()
        ser = ArrowStreamPandasSerializer(timezone, safecheck, False)

        @no_type_check
        def reader_func(temp_filename):
            return self._jvm.PythonSQLUtils.readArrowStreamFromFile(temp_filename)

        @no_type_check
        def create_iter_server():
            return self._jvm.ArrowIteratorServer()

        # Create Spark DataFrame from Arrow stream file, using one batch per partition
        jiter = self._sc._serialize_to_jvm(arrow_data, ser, reader_func, create_iter_server)
        assert self._jvm is not None
        jdf = self._jvm.PythonSQLUtils.toDataFrame(jiter, schema.json(), jsparkSession)
        df = DataFrame(jdf, self)
        df._schema = schema
        return df

    def _create_from_arrow_table(
        self, table: "pa.Table", schema: Union[StructType, List[str]], timezone: str
    ) -> "DataFrame":
        """
        Create a DataFrame from a given pyarrow.Table by slicing it into partitions then
        sending to the JVM to parallelize.
        """
        from pyspark.sql import SparkSession
        from pyspark.sql.dataframe import DataFrame

        assert isinstance(self, SparkSession)

        from pyspark.sql.pandas.serializers import ArrowStreamSerializer
        from pyspark.sql.pandas.types import (
            from_arrow_type,
            from_arrow_schema,
            to_arrow_schema,
            _check_arrow_table_timestamps_localize,
        )
        from pyspark.sql.pandas.utils import require_minimum_pyarrow_version

        require_minimum_pyarrow_version()

        prefer_timestamp_ntz = is_timestamp_ntz_preferred()

        # Create the Spark schema from list of names passed in with Arrow types
        if isinstance(schema, (list, tuple)):
            table = table.rename_columns(schema)
            arrow_schema = table.schema
            struct = StructType()
            for name, field in zip(schema, arrow_schema):
                struct.add(
                    name,
                    from_arrow_type(field.type, prefer_timestamp_ntz),
                    nullable=field.nullable,
                )
            schema = struct

        if not isinstance(schema, StructType):
            schema = from_arrow_schema(table.schema, prefer_timestamp_ntz=prefer_timestamp_ntz)

        prefers_large_var_types = self._jconf.arrowUseLargeVarTypes()
        table = _check_arrow_table_timestamps_localize(table, schema, True, timezone).cast(
            to_arrow_schema(
                schema,
                error_on_duplicated_field_names_in_struct=True,
                prefers_large_types=prefers_large_var_types,
            )
        )

        # Chunk the Arrow Table into RecordBatches
        chunk_size = self._jconf.arrowMaxRecordsPerBatch()
        arrow_data = table.to_batches(max_chunksize=chunk_size)

        jsparkSession = self._jsparkSession

        ser = ArrowStreamSerializer()

        @no_type_check
        def reader_func(temp_filename):
            return self._jvm.PythonSQLUtils.readArrowStreamFromFile(temp_filename)

        @no_type_check
        def create_iter_server():
            return self._jvm.ArrowIteratorServer()

        # Create Spark DataFrame from Arrow stream file, using one batch per partition
        jiter = self._sc._serialize_to_jvm(arrow_data, ser, reader_func, create_iter_server)
        assert self._jvm is not None
        jdf = self._jvm.PythonSQLUtils.toDataFrame(jiter, schema.json(), jsparkSession)
        df = DataFrame(jdf, self)
        df._schema = schema
        return df


def _test() -> None:
    import doctest
    from pyspark.sql import SparkSession
    import pyspark.sql.pandas.conversion

    globs = pyspark.sql.pandas.conversion.__dict__.copy()
    spark = (
        SparkSession.builder.master("local[4]").appName("sql.pandas.conversion tests").getOrCreate()
    )
    globs["spark"] = spark
    (failure_count, test_count) = doctest.testmod(
        pyspark.sql.pandas.conversion,
        globs=globs,
        optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE | doctest.REPORT_NDIFF,
    )
    spark.stop()
    if failure_count:
        sys.exit(-1)


if __name__ == "__main__":
    _test()
